#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "d88image.h"


struct ram_data_info_t {
	const char *filename;
	int start_track;
	int end_track;
};


static void usage(void);
static int retrieve_ram(const struct ram_data_info_t *ram_data_info, const char *image_file, int image_index, int overwrite);
static int save_track_data(const struct d88_disk_t *disk, int start_track, int num_tracks, const char *filename, int overwrite);
static int confirm(const char *msg);


int main(int argc, char **argv)
{
	int num_cmd_params = 1;
	const char *image_file = NULL;
	int image_index = 0;
	struct ram_data_info_t ram_data_info[6] = {
		{ "main_ram.bin", 2, 17 },
		{ "high_speed_ram.bin", 18, 18 },
		{ "ext_ram.bin", 19, 50 }
	};
	int save_ext_ram = 0;
	static int overwrite = 0;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-i") == 0 || strcmp(*argv, "--index") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("error: no image index\n");
					return 1;
				}
				image_index = strtoul(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0') {
					printf("error: illegal image index '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "-m") == 0 || strcmp(*argv, "--main") == 0) {
				if (*++argv == NULL) {
					printf("error: no main ram filename\n");
					return 1;
				}
				ram_data_info[0].filename = *argv;
			} else if (strcmp(*argv, "-h") == 0 || strcmp(*argv, "--high") == 0) {
				if (*++argv == NULL) {
					printf("error: no high speed ram filename\n");
					return 1;
				}
				ram_data_info[1].filename = *argv;
			} else if (strcmp(*argv, "-H") == 0 || strcmp(*argv, "--no-high") == 0) {
				ram_data_info[1].filename = NULL;
			} else if (strcmp(*argv, "-E") == 0) {
				save_ext_ram = 1;
			} else if (strcmp(*argv, "-e") == 0 || strcmp(*argv, "--ext") == 0) {
				if (*++argv == NULL) {
					printf("error: no ext ram filename\n");
					return 1;
				}
				ram_data_info[2].filename = *argv;
				save_ext_ram = 1;
			} else if (strcmp(*argv, "--overwrite") == 0) {
				overwrite = 1;
			} else {
				printf("error: unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("error: illegal parameter '%s'\n", *argv);
				return 1;
			} else {
				image_file = *argv;
				num_cmd_params--;
			}
		}
	}

	if (image_file == NULL) {
		printf("error: no d88 image filename\n");
		return 1;
	}

	if (save_ext_ram == 0) {
		ram_data_info[2].filename = NULL;
	}
	if (retrieve_ram(ram_data_info, image_file, image_index, overwrite) < 0) {
		printf("error: can't retrieve ram data\n");
		return 1;
	}

	return 0;
}

static void usage(void)
{
	printf("usage: retrieve_ram [options] D88-FILE\n");
	printf("option: -i|--index NUM      specify image index\n");
	printf("        -m|--main FILENAME  write main ram data to specified file\n");
	printf("        -h|--high FILENAME  write high speed ram data to specified file\n");
	printf("        -H|--no-high        not retrieve high speed ram data\n");
	printf("        -E|                 retrieve ext ram data\n");
	printf("        -e|--ext FILENAME   write ext ram data to specified file\n");
	printf("           --overwrite      overwrite file(s)\n");
	printf("        -?|--help           print usage\n");
}

static int retrieve_ram(const struct ram_data_info_t *ram_data_info, const char *image_file, int image_index, int overwrite)
{
	int result = -1;
	struct d88_image_t image;
	struct d88_disk_t *disk = NULL;
	int i;

	memset(&image, 0, sizeof(image));

	if (read_d88_image(&image, image_file) < 0) {
		printf("error: read d88 image '%s'\n", image_file);
		goto ERROR_EXIT;
	}
	if ((disk = get_d88_disk(&image, image_index)) == NULL) {
		printf("error: get disk %d from image '%s'\n", image_index, image_file);
		goto ERROR_EXIT;
	}

	for (i = 0; i < 3; i++) {
		if (ram_data_info[i].filename == NULL) {
			continue;
		}

		if (save_track_data(disk, ram_data_info[i].start_track, ram_data_info[i].end_track, ram_data_info[i].filename, overwrite) < 0) {
			printf("error: retrieve main ram data from '%s':%d\n", image_file, image_index);
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	free_d88_image(&image, 0);

	return result;
}

static int save_track_data(const struct d88_disk_t *disk, int start_track, int end_track, const char *filename, int overwrite)
{
	int result = -1;
	FILE *fp = NULL;
	struct stat st;
	int trk;
	int sec;

	if (D88_TRACK_COUNT <= end_track) {
		printf("error: illegal track no: %d - %d\n", start_track, end_track);
		goto ERROR_EXIT;
	}

	if (stat(filename, &st) == 0) {
		printf("already exists '%s' ... overwrite", filename);
		if (overwrite) {
			printf("\n");
		} else if (confirm(" ?") <= 0) {
			goto ERROR_EXIT;
		}
	}

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("error: can't open '%s'\n", filename);
		goto ERROR_EXIT;
	}

	for (trk = start_track; trk <= end_track; trk++) {
		if (disk->tracks[trk].num_sectors != 16) {
			printf("error: illegal track format: track=%d, num sectors=%d\n", trk, disk->tracks[trk].num_sectors);
			goto ERROR_EXIT;
		}

		for (sec = 0; sec < disk->tracks[trk].num_sectors; sec++) {
			const unsigned char *data = disk->tracks[trk].sectors[sec].data;

			if (disk->tracks[trk].sectors[sec].sector_size != 256) {
				printf("error: illegal sector size: track=%d, sector=%d, size=%d\n", trk, sec, disk->tracks[trk].sectors[sec].sector_size);
				goto ERROR_EXIT;
			}

			if (fwrite(data, 256, 1, fp) != 1) {
				printf("error: can't write '%s' track=%d, sector=%d\n", filename, trk, sec);
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
